#include "includes/Definitions.hpp"

#include "Sphere.cpp"
#include "CoreShell.cpp"
#include "Cylinder.cpp"
#include "Sources.cpp"
#include "Detectors.cpp"




  class Experiment
  {
  public:
      SPHERE::Set         sphereSet;
      CYLINDER::Set       cylinderSet;
      CORESHELL::Set      coreshellSet;
      DETECTOR::Set       detectorSet;
      SOURCE::Set         sourceSet;

      Experiment(){}


void SetSphere(SPHERE::Set& ScattererSet){ this->sphereSet = ScattererSet;}

void SetCylinder(CYLINDER::Set& ScattererSet) { this->cylinderSet = ScattererSet; }

void SetCoreShell(CORESHELL::Set& ScattererSet) { this->coreshellSet = ScattererSet; }

void SetSource( SOURCE::Set &sourceSet){ this->sourceSet = sourceSet; }

void SetDetector( DETECTOR::Set &detectorSet){ this->detectorSet = detectorSet; }



//--------------------------------------SPHERE------------------------------------
//--------------------------------------SPHERE------------------------------------
Cndarray GetSphereCoefficient(CVector (SPHERE::Scatterer::*function)(void), size_t MaxOrder=0)
{
  if (sphereSet.BoundedIndex)
      return GetSphereCoefficientBound0(function, MaxOrder);
  else
      return GetSphereCoefficientUnbound(function, MaxOrder);
}

ndarray GetSphereData(double (SPHERE::Scatterer::*function)(void), size_t MaxOrder=0)
{
  if (sphereSet.BoundedIndex)
      return GetSphereDataBound0(function, MaxOrder);
  else
      return GetSphereDataUnbound(function, MaxOrder);
}


ndarray GetSphereCoupling()
{
  if (sphereSet.BoundedIndex)
      return GetSphereCouplingBound0();
  else
      return GetSphereCouplingUnbound();
}




Cndarray GetSphereCoefficientBound0(CVector (SPHERE::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace SPHERE;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   sphereSet.Diameter.size(),
                                   sphereSet.Material.size(),
                                   sphereSet.nMedium.size()};


  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  CVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          {
                            size_t idx = n  +
                                         i  * FullShape[5] +
                                         d  * FullShape[5] * FullShape[4] +
                                         a  * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            SPHERE::State   scattererState = SPHERE::State(sphereSet.Diameter[d], sphereSet.Material[i][w], sphereSet.nMedium[n]);

                            SPHERE::Scatterer Scat = SPHERE::Scatterer(scattererState, sourceState, MaxOrder+1);

                            Output[idx] = (Scat.*function)()[MaxOrder];
                          }

  return Vector2ndarray(Output, FullShape);
}






Cndarray GetSphereCoefficientUnbound(CVector (SPHERE::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace SPHERE;

  std::vector<size_t> FullShape;

  FullShape = {sourceSet.Wavelength.size(),
               sourceSet.Jones.size(),
               sourceSet.Amplitude.size(),
               sphereSet.Diameter.size(),
               sphereSet.Index.size(),
               sphereSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  CVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          {
                            size_t idx = n  +
                                         i  * FullShape[5] +
                                         d  * FullShape[5] * FullShape[4] +
                                         a  * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            SPHERE::State   scattererState = SPHERE::State(sphereSet.Diameter[d], sphereSet.Index[i], sphereSet.nMedium[n]);

                            SPHERE::Scatterer Scat = SPHERE::Scatterer(scattererState, sourceState, MaxOrder+1);

                            Output[idx] = (Scat.*function)()[MaxOrder];
                          }

  return Vector2ndarray(Output, FullShape);
}






ndarray GetSphereDataBound0(double (SPHERE::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace SPHERE;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   sphereSet.Diameter.size(),
                                   sphereSet.Material.size(),
                                   sphereSet.nMedium.size()};


  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  DVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          {
                            size_t idx = n  +
                                         i  * FullShape[5] +
                                         d  * FullShape[5] * FullShape[4] +
                                         a  * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            SPHERE::State   scattererState = SPHERE::State(sphereSet.Diameter[d], sphereSet.Material[i][w], sphereSet.nMedium[n]);

                            SPHERE::Scatterer Scat = SPHERE::Scatterer(scattererState, sourceState);

                            Output[idx] = (Scat.*function)();
                          }

  return Vector2ndarray(Output, FullShape);
}






ndarray GetSphereDataUnbound(double (SPHERE::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace SPHERE;

  std::vector<size_t> FullShape;

  FullShape = {sourceSet.Wavelength.size(),
               sourceSet.Jones.size(),
               sourceSet.Amplitude.size(),
               sphereSet.Diameter.size(),
               sphereSet.Index.size(),
               sphereSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          {
                            size_t idx = n  +
                                         i  * FullShape[5] +
                                         d  * FullShape[5] * FullShape[4] +
                                         a  * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            SPHERE::State   scattererState = SPHERE::State(sphereSet.Diameter[d], sphereSet.Index[i], sphereSet.nMedium[n]);

                            SPHERE::Scatterer Scat = SPHERE::Scatterer(scattererState, sourceState, MaxOrder);

                            Output[idx] = (Scat.*function)();
                          }

  return Vector2ndarray(Output, FullShape);
}






ndarray GetSphereCouplingBound0()
{
  using namespace SPHERE;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   sphereSet.Diameter.size(),
                                   sphereSet.Material.size(),
                                   sphereSet.nMedium.size(),
                                   detectorSet.ScalarField.size(),
                                   detectorSet.NA.size(),
                                   detectorSet.PhiOffset.size(),
                                   detectorSet.GammaOffset.size(),
                                   detectorSet.Filter.size()};


  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(11)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          for (size_t s=0; s<FullShape[6]; ++s)
                              for (size_t na=0; na<FullShape[7]; ++na)
                                  for (size_t p=0; p<FullShape[8]; ++p)
                                      for (size_t g=0; g<FullShape[9]; ++g)
                                          for (size_t f=0; f<FullShape[10]; ++f)
                                        {
                                          size_t idx = f  +
                                                       g  * FullShape[10] +
                                                       p  * FullShape[10] * FullShape[9] +
                                                       na * FullShape[10] * FullShape[9] * FullShape[8] +
                                                       s  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] +
                                                       n  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] +
                                                       i  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] +
                                                       d  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                                       a  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                                       j  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                                       w  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                                       ;

                                          SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                          SPHERE::State   scattererState = SPHERE::State(sphereSet.Diameter[d], sphereSet.Material[i][w], sphereSet.nMedium[n]);
                                          DETECTOR::State detectorState  = DETECTOR::State(detectorSet.ScalarField[s], detectorSet.NA[na], detectorSet.PhiOffset[p], detectorSet.GammaOffset[g], detectorSet.Filter[f], detectorSet.Coherent, detectorSet.PointCoupling);

                                          SPHERE::Scatterer Scat = SPHERE::Scatterer(scattererState, sourceState);
                                          DETECTOR::Detector det = DETECTOR::Detector(detectorState);

                                          Output[idx] = abs( det.Coupling(Scat) );
                                        }

  return Vector2ndarray(Output, FullShape);
}






ndarray GetSphereCouplingUnbound()
{
  using namespace SPHERE;

  std::vector<size_t> FullShape;

  FullShape = {sourceSet.Wavelength.size(),
               sourceSet.Jones.size(),
               sourceSet.Amplitude.size(),
               sphereSet.Diameter.size(),
               sphereSet.Index.size(),
               sphereSet.nMedium.size(),
               detectorSet.ScalarField.size(),
               detectorSet.NA.size(),
               detectorSet.PhiOffset.size(),
               detectorSet.GammaOffset.size(),
               detectorSet.Filter.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(11)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          for (size_t s=0; s<FullShape[6]; ++s)
                              for (size_t na=0; na<FullShape[7]; ++na)
                                  for (size_t p=0; p<FullShape[8]; ++p)
                                      for (size_t g=0; g<FullShape[9]; ++g)
                                          for (size_t f=0; f<FullShape[10]; ++f)
                                        {
                                          size_t idx = f  +
                                                       g  * FullShape[10] +
                                                       p  * FullShape[10] * FullShape[9] +
                                                       na * FullShape[10] * FullShape[9] * FullShape[8] +
                                                       s  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] +
                                                       n  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] +
                                                       i  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] +
                                                       d  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                                       a  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                                       j  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                                       w  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1]
                                                       ;



                                          SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                          SPHERE::State   scattererState = SPHERE::State(sphereSet.Diameter[d], sphereSet.Index[i], sphereSet.nMedium[n]);
                                          DETECTOR::State detectorState  = DETECTOR::State(detectorSet.ScalarField[s], detectorSet.NA[na], detectorSet.PhiOffset[p], detectorSet.GammaOffset[g], detectorSet.Filter[f], detectorSet.Coherent, detectorSet.PointCoupling);

                                          SPHERE::Scatterer Scat = SPHERE::Scatterer(scattererState, sourceState);
                                          DETECTOR::Detector det = DETECTOR::Detector(detectorState);

                                          Output[idx] = abs( det.Coupling(Scat) );
                                        }

  return Vector2ndarray(Output, FullShape);
}

















//--------------------------------------CYLINDER------------------------------------
//--------------------------------------CYLINDER------------------------------------
Cndarray GetCylinderCoefficient(CVector (CYLINDER::Scatterer::*function)(void), size_t MaxOrder=0)
{
  if (cylinderSet.BoundedIndex)
      return GetCylinderCoefficientBound0(function, MaxOrder);
  else
      return GetCylinderCoefficientUnbound(function, MaxOrder);
}

ndarray GetCylinderData(double (CYLINDER::Scatterer::*function)(void), size_t MaxOrder=0)
{
  if (cylinderSet.BoundedIndex)
      return GetCylinderDataBound0(function, MaxOrder);
  else
      return GetCylinderDataUnbound(function, MaxOrder);
}

ndarray GetCylinderCoupling()
{
  if (cylinderSet.BoundedIndex)
      return GetCylinderCouplingBound0();
  else
      return GetCylinderCouplingUnbound();
}



Cndarray GetCylinderCoefficientBound0(CVector (CYLINDER::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CYLINDER;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   cylinderSet.Diameter.size(),
                                   cylinderSet.Material.size(),
                                   cylinderSet.nMedium.size()};


  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  CVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          {
                            size_t idx = n  +
                                         i  * FullShape[5] +
                                         d  * FullShape[5] * FullShape[4] +
                                         a  * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CYLINDER::State scattererState = CYLINDER::State(cylinderSet.Diameter[d], cylinderSet.Material[i][w], cylinderSet.nMedium[n]);

                            CYLINDER::Scatterer Scat = CYLINDER::Scatterer(scattererState, sourceState, MaxOrder+1);

                            Output[idx] = (Scat.*function)()[MaxOrder];
                          }

  return Vector2ndarray(Output, FullShape);
}






Cndarray GetCylinderCoefficientUnbound(CVector (CYLINDER::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CYLINDER;

  std::vector<size_t> FullShape;

  FullShape = {sourceSet.Wavelength.size(),
               sourceSet.Jones.size(),
               sourceSet.Amplitude.size(),
               cylinderSet.Diameter.size(),
               cylinderSet.Index.size(),
               cylinderSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  CVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          {
                            size_t idx = n  +
                                         i  * FullShape[5] +
                                         d  * FullShape[5] * FullShape[4] +
                                         a  * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CYLINDER::State scattererState = CYLINDER::State(cylinderSet.Diameter[d], cylinderSet.Index[i], cylinderSet.nMedium[n]);

                            CYLINDER::Scatterer Scat = CYLINDER::Scatterer(scattererState, sourceState, MaxOrder+1);

                            Output[idx] = (Scat.*function)()[MaxOrder];
                          }

  return Vector2ndarray(Output, FullShape);
}



ndarray GetCylinderDataBound0(double (CYLINDER::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CYLINDER;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   cylinderSet.Diameter.size(),
                                   cylinderSet.Material.size(),
                                   cylinderSet.nMedium.size()};


  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  DVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          {
                            size_t idx = n  +
                                         i  * FullShape[5] +
                                         d  * FullShape[5] * FullShape[4] +
                                         a  * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CYLINDER::State scattererState = CYLINDER::State(cylinderSet.Diameter[d], cylinderSet.Material[i][w], cylinderSet.nMedium[n]);

                            CYLINDER::Scatterer Scat = CYLINDER::Scatterer(scattererState, sourceState, MaxOrder);

                            Output[idx] = (Scat.*function)();
                          }

  return Vector2ndarray(Output, FullShape);
}






ndarray GetCylinderDataUnbound(double (CYLINDER::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CYLINDER;

  std::vector<size_t> FullShape;

  FullShape = {sourceSet.Wavelength.size(),
               sourceSet.Jones.size(),
               sourceSet.Amplitude.size(),
               cylinderSet.Diameter.size(),
               cylinderSet.Index.size(),
               cylinderSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          {
                            size_t idx = n  +
                                         i  * FullShape[5] +
                                         d  * FullShape[5] * FullShape[4] +
                                         a  * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CYLINDER::State scattererState = CYLINDER::State(cylinderSet.Diameter[d], cylinderSet.Index[i], cylinderSet.nMedium[n]);

                            CYLINDER::Scatterer Scat = CYLINDER::Scatterer(scattererState, sourceState, MaxOrder);

                            Output[idx] = (Scat.*function)();
                          }

  return Vector2ndarray(Output, FullShape);
}







ndarray GetCylinderCouplingBound0()
{
  using namespace CYLINDER;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   cylinderSet.Diameter.size(),
                                   cylinderSet.Material.size(),
                                   cylinderSet.nMedium.size(),
                                   detectorSet.ScalarField.size(),
                                   detectorSet.NA.size(),
                                   detectorSet.PhiOffset.size(),
                                   detectorSet.GammaOffset.size(),
                                   detectorSet.Filter.size()};


  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(11)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          for (size_t s=0; s<FullShape[6]; ++s)
                              for (size_t na=0; na<FullShape[7]; ++na)
                                  for (size_t p=0; p<FullShape[8]; ++p)
                                      for (size_t g=0; g<FullShape[9]; ++g)
                                          for (size_t f=0; f<FullShape[10]; ++f)
                                        {
                                          size_t idx = f  +
                                                       g  * FullShape[10] +
                                                       p  * FullShape[10] * FullShape[9] +
                                                       na * FullShape[10] * FullShape[9] * FullShape[8] +
                                                       s  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] +
                                                       n  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] +
                                                       i  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] +
                                                       d  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                                       a  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                                       j  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                                       w  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                                       ;

                                          SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                          CYLINDER::State scattererState = CYLINDER::State(cylinderSet.Diameter[d], cylinderSet.Material[i][w], cylinderSet.nMedium[n]);
                                          DETECTOR::State detectorState  = DETECTOR::State(detectorSet.ScalarField[s], detectorSet.NA[na], detectorSet.PhiOffset[p], detectorSet.GammaOffset[g], detectorSet.Filter[f], detectorSet.Coherent, detectorSet.PointCoupling);

                                          CYLINDER::Scatterer Scat = CYLINDER::Scatterer(scattererState, sourceState);
                                          DETECTOR::Detector  det = DETECTOR::Detector(detectorState);

                                          Output[idx] = abs( det.Coupling(Scat) );
                                        }

  return Vector2ndarray(Output, FullShape);
}






ndarray GetCylinderCouplingUnbound()
{
  using namespace CYLINDER;

  std::vector<size_t> FullShape;

  FullShape = {sourceSet.Wavelength.size(),
               sourceSet.Jones.size(),
               sourceSet.Amplitude.size(),
               cylinderSet.Diameter.size(),
               cylinderSet.Index.size(),
               cylinderSet.nMedium.size(),
               detectorSet.ScalarField.size(),
               detectorSet.NA.size(),
               detectorSet.PhiOffset.size(),
               detectorSet.GammaOffset.size(),
               detectorSet.Filter.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(11)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t d=0; d<FullShape[3]; ++d)
                  for (size_t i=0; i<FullShape[4]; ++i)
                      for (size_t n=0; n<FullShape[5]; ++n)
                          for (size_t s=0; s<FullShape[6]; ++s)
                              for (size_t na=0; na<FullShape[7]; ++na)
                                  for (size_t p=0; p<FullShape[8]; ++p)
                                      for (size_t g=0; g<FullShape[9]; ++g)
                                          for (size_t f=0; f<FullShape[10]; ++f)
                                        {
                                          size_t idx = f  +
                                                       g  * FullShape[10] +
                                                       p  * FullShape[10] * FullShape[9] +
                                                       na * FullShape[10] * FullShape[9] * FullShape[8] +
                                                       s  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] +
                                                       n  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] +
                                                       i  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] +
                                                       d  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                                       a  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                                       j  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                                       w  * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                                       ;

                                          SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                          CYLINDER::State scattererState = CYLINDER::State(cylinderSet.Diameter[d], cylinderSet.Index[i], cylinderSet.nMedium[n]);
                                          DETECTOR::State detectorState  = DETECTOR::State(detectorSet.ScalarField[s], detectorSet.NA[na], detectorSet.PhiOffset[p], detectorSet.GammaOffset[g], detectorSet.Filter[f], detectorSet.Coherent, detectorSet.PointCoupling);

                                          CYLINDER::Scatterer Scat = CYLINDER::Scatterer(scattererState, sourceState);
                                          DETECTOR::Detector  det = DETECTOR::Detector(detectorState);

                                          Output[idx] = abs( det.Coupling(Scat) );
                                        }

  return Vector2ndarray(Output, FullShape);
}







//--------------------------------------CORESHELL------------------------------------
//--------------------------------------CORESHELL------------------------------------
Cndarray GetCoreShellCoefficient(CVector (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{

  if (coreshellSet.BoundedCore && coreshellSet.BoundedShell)
      return GetCoreShellCoefficientBoundCoreBoundShell(function, MaxOrder);

  if (coreshellSet.BoundedCore && ~coreshellSet.BoundedShell)
      return GetCoreShellCoefficientBoundCore(function, MaxOrder);

  if (~coreshellSet.BoundedCore && coreshellSet.BoundedShell)
      return GetCoreShellCoefficientBoundShell(function, MaxOrder);

  if (~coreshellSet.BoundedCore && ~coreshellSet.BoundedShell)
      return GetCoreShellCoefficientUnbound(function, MaxOrder);

}

ndarray GetCoreShellData(double (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  if (coreshellSet.BoundedCore && coreshellSet.BoundedShell)
      return GetCoreShellDataBoundCoreBoundShell(function, MaxOrder);

  if (coreshellSet.BoundedCore && ~coreshellSet.BoundedShell)
      return GetCoreShellDataBoundCore(function, MaxOrder);

  if (~coreshellSet.BoundedCore && coreshellSet.BoundedShell)
      return GetCoreShellDataBoundShell(function, MaxOrder);

  if (~coreshellSet.BoundedCore && ~coreshellSet.BoundedShell)
      return GetCoreShellDataUnbound(function, MaxOrder);

}


ndarray GetCoreShellCoupling()
{
  if (coreshellSet.BoundedCore && coreshellSet.BoundedShell)
      return GetCoreShellCouplingBoundCoreBoundShell();

  if (coreshellSet.BoundedCore && ~coreshellSet.BoundedShell)
      return GetCoreShellCouplingBoundCore();

  if (~coreshellSet.BoundedCore && coreshellSet.BoundedShell)
      return GetCoreShellCouplingBoundShell();

  if (~coreshellSet.BoundedCore && ~coreshellSet.BoundedShell)
  {
    return GetCoreShellCouplingUnbound();
  }

}


Cndarray GetCoreShellCoefficientBoundCore(CVector (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreMaterial.size(),
                                   coreshellSet.ShellIndex.size(),
                                   coreshellSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  CVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                          {
                            size_t idx = n  +
                                         Si * FullShape[7] +
                                         Ci * FullShape[7] * FullShape[6] +
                                         Sd * FullShape[7] * FullShape[6] * FullShape[5] +
                                         Cd * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                         a  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[2] * FullShape[1]
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CORESHELL::State scattererState = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreMaterial[Ci][w], coreshellSet.ShellIndex[Si], coreshellSet.nMedium[n]);

                            CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState, MaxOrder+1);

                            Output[idx] = (Scat.*function)()[MaxOrder];
                          }


  return Vector2ndarray(Output, FullShape);
}



Cndarray GetCoreShellCoefficientBoundShell(CVector (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreIndex.size(),
                                   coreshellSet.ShellMaterial.size(),
                                   coreshellSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  CVector Output(FullSize);

  #pragma omp parallel for collapse(8)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                          {
                            size_t idx = n  +
                                         Si * FullShape[7] +
                                         Ci * FullShape[7] * FullShape[6] +
                                         Sd * FullShape[7] * FullShape[6] * FullShape[5] +
                                         Cd * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                         a  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[2] * FullShape[1]
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CORESHELL::State scattererState = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreIndex[Ci], coreshellSet.ShellMaterial[Si][w], coreshellSet.nMedium[n]);

                            CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState, MaxOrder+1);

                            Output[idx] = (Scat.*function)()[MaxOrder];
                          }


  return Vector2ndarray(Output, FullShape);
}



Cndarray GetCoreShellCoefficientBoundCoreBoundShell(CVector (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreMaterial.size(),
                                   coreshellSet.ShellMaterial.size(),
                                   coreshellSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  CVector Output(FullSize);

  #pragma omp parallel for collapse(8)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                          {
                            size_t idx = n  +
                                         Si * FullShape[7] +
                                         Ci * FullShape[7] * FullShape[6] +
                                         Sd * FullShape[7] * FullShape[6] * FullShape[5] +
                                         Cd * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                         a  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[2] * FullShape[1]
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CORESHELL::State scattererState = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreMaterial[Ci][w], coreshellSet.ShellMaterial[Si][w], coreshellSet.nMedium[n]);

                            CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState, MaxOrder+1);

                            Output[idx] = (Scat.*function)()[MaxOrder];
                          }

  return Vector2ndarray(Output, FullShape);
}




Cndarray GetCoreShellCoefficientUnbound(CVector (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape;

  FullShape = {sourceSet.Wavelength.size(),
               sourceSet.Jones.size(),
               sourceSet.Amplitude.size(),
               coreshellSet.CoreDiameter.size(),
               coreshellSet.ShellDiameter.size(),
               coreshellSet.CoreIndex.size(),
               coreshellSet.ShellIndex.size(),
               coreshellSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  CVector Output(FullSize);

  #pragma omp parallel for collapse(8)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                              {
                                size_t idx = n  +
                                             Si * FullShape[7] +
                                             Ci * FullShape[7] * FullShape[6] +
                                             Sd * FullShape[7] * FullShape[6] * FullShape[5] +
                                             Cd * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                             a  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                             j  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                             w  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                             ;

                                SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                CORESHELL::State scattererState = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreIndex[Ci], coreshellSet.ShellIndex[Si], coreshellSet.nMedium[n]);

                                CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState, MaxOrder+1);

                                Output[idx] = (Scat.*function)()[MaxOrder];
                              }

  return Vector2ndarray(Output, FullShape);
}





ndarray GetCoreShellDataBoundCore(double (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreMaterial.size(),
                                   coreshellSet.ShellIndex.size(),
                                   coreshellSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  DVector Output(FullSize);

  #pragma omp parallel for collapse(6)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                          {
                            size_t idx = n  +
                                         Si * FullShape[7] +
                                         Ci * FullShape[7] * FullShape[6] +
                                         Sd * FullShape[7] * FullShape[6] * FullShape[5] +
                                         Cd * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                         a  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[2] * FullShape[1]
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CORESHELL::State scattererState = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreMaterial[Ci][w], coreshellSet.ShellIndex[Si], coreshellSet.nMedium[n]);

                            CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState, MaxOrder);

                            Output[idx] = (Scat.*function)();
                          }


  return Vector2ndarray(Output, FullShape);
}



ndarray GetCoreShellDataBoundShell(double (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreIndex.size(),
                                   coreshellSet.ShellMaterial.size(),
                                   coreshellSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  DVector Output(FullSize);

  #pragma omp parallel for collapse(8)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                          {
                            size_t idx = n  +
                                         Si * FullShape[7] +
                                         Ci * FullShape[7] * FullShape[6] +
                                         Sd * FullShape[7] * FullShape[6] * FullShape[5] +
                                         Cd * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                         a  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[2] * FullShape[1]
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CORESHELL::State scattererState = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreIndex[Ci], coreshellSet.ShellMaterial[Si][w], coreshellSet.nMedium[n]);

                            CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState, MaxOrder);

                            Output[idx] = (Scat.*function)();
                          }


  return Vector2ndarray(Output, FullShape);
}



ndarray GetCoreShellDataBoundCoreBoundShell(double (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreMaterial.size(),
                                   coreshellSet.ShellMaterial.size(),
                                   coreshellSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  DVector Output(FullSize);

  #pragma omp parallel for collapse(8)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                          {
                            size_t idx = n  +
                                         Si * FullShape[7] +
                                         Ci * FullShape[7] * FullShape[6] +
                                         Sd * FullShape[7] * FullShape[6] * FullShape[5] +
                                         Cd * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                         a  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                         j  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                         w  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[2] * FullShape[1]
                                         ;

                            SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                            CORESHELL::State scattererState = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreMaterial[Ci][w], coreshellSet.ShellMaterial[Si][w], coreshellSet.nMedium[n]);

                            CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState, MaxOrder);

                            Output[idx] = (Scat.*function)();
                          }

  return Vector2ndarray(Output, FullShape);
}




ndarray GetCoreShellDataUnbound(double (CORESHELL::Scatterer::*function)(void), size_t MaxOrder=0)
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape;

  FullShape = {sourceSet.Wavelength.size(),
               sourceSet.Jones.size(),
               sourceSet.Amplitude.size(),
               coreshellSet.CoreDiameter.size(),
               coreshellSet.ShellDiameter.size(),
               coreshellSet.CoreIndex.size(),
               coreshellSet.ShellIndex.size(),
               coreshellSet.nMedium.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(8)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                              {
                                size_t idx = n  +
                                             Si * FullShape[7] +
                                             Ci * FullShape[7] * FullShape[6] +
                                             Sd * FullShape[7] * FullShape[6] * FullShape[5] +
                                             Cd * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                             a  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                             j  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                             w  * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                             ;

                                SOURCE::State   sourceState    = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                CORESHELL::State scattererState = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreIndex[Ci], coreshellSet.ShellIndex[Si], coreshellSet.nMedium[n]);

                                CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState, MaxOrder);

                                Output[idx] = (Scat.*function)();
                              }

  return Vector2ndarray(Output, FullShape);
}











ndarray GetCoreShellCouplingUnbound()
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreIndex.size(),
                                   coreshellSet.ShellIndex.size(),
                                   coreshellSet.nMedium.size(),
                                   detectorSet.ScalarField.size(),
                                   detectorSet.NA.size(),
                                   detectorSet.PhiOffset.size(),
                                   detectorSet.GammaOffset.size(),
                                   detectorSet.Filter.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);


  #pragma omp parallel for collapse(13)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                                  for (size_t s=0; s<FullShape[8]; ++s)
                                      for (size_t na=0; na<FullShape[9]; ++na)
                                          for (size_t p=0; p<FullShape[10]; ++p)
                                              for (size_t g=0; g<FullShape[11]; ++g)
                                                  for (size_t f=0; f<FullShape[12]; ++f)
                                                  {
                                                    size_t idx = f  +
                                                                 g  * FullShape[12] +
                                                                 p  * FullShape[12] * FullShape[11] +
                                                                 na * FullShape[12] * FullShape[11] * FullShape[10] +
                                                                 s  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] +
                                                                 n  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] +
                                                                 Si * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] +
                                                                 Ci * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] +
                                                                 Sd * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] +
                                                                 Cd * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                                                 a  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                                                 j  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                                                 w  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1]
                                                                 ;

                                                    SOURCE::State    sourceState     = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                                    CORESHELL::State scattererState  = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreIndex[Ci], coreshellSet.ShellIndex[Si], coreshellSet.nMedium[n]);
                                                    DETECTOR::State  detectorState   = DETECTOR::State(detectorSet.ScalarField[s], detectorSet.NA[na], detectorSet.PhiOffset[p], detectorSet.GammaOffset[g], detectorSet.Filter[f], detectorSet.Coherent, detectorSet.PointCoupling);

                                                    CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState);
                                                    DETECTOR::Detector det = DETECTOR::Detector(detectorState);

                                                    Output[idx] = abs( det.Coupling(Scat) );
                                                  }

  return Vector2ndarray(Output, FullShape);
}




ndarray GetCoreShellCouplingBoundCore()
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreMaterial.size(),
                                   coreshellSet.ShellIndex.size(),
                                   coreshellSet.nMedium.size(),
                                   detectorSet.ScalarField.size(),
                                   detectorSet.NA.size(),
                                   detectorSet.PhiOffset.size(),
                                   detectorSet.GammaOffset.size(),
                                   detectorSet.Filter.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(13)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                                  for (size_t s=0; s<FullShape[8]; ++s)
                                      for (size_t na=0; na<FullShape[9]; ++na)
                                          for (size_t p=0; p<FullShape[10]; ++p)
                                              for (size_t g=0; g<FullShape[11]; ++g)
                                                  for (size_t f=0; f<FullShape[12]; ++f)
                                                  {
                                                    size_t idx = f  +
                                                                 g  * FullShape[12] +
                                                                 p  * FullShape[12] * FullShape[11] +
                                                                 na * FullShape[12] * FullShape[11] * FullShape[10] +
                                                                 s  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] +
                                                                 n  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] +
                                                                 Si * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] +
                                                                 Ci * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] +
                                                                 Sd * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] +
                                                                 Cd * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                                                 a  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                                                 j  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                                                 w  * FullShape[12] * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1]
                                                                 ;

                                                    SOURCE::State    sourceState     = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                                    CORESHELL::State scattererState  = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreMaterial[Ci][w], coreshellSet.ShellIndex[Si], coreshellSet.nMedium[n]);
                                                    DETECTOR::State  detectorState   = DETECTOR::State(detectorSet.ScalarField[0], detectorSet.NA[na], detectorSet.PhiOffset[p], detectorSet.GammaOffset[g], detectorSet.Filter[f], detectorSet.Coherent, detectorSet.PointCoupling);

                                                    CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState);
                                                    DETECTOR::Detector det = DETECTOR::Detector(detectorState);

                                                    Output[idx] = abs( det.Coupling(Scat) );
                                                  }

  return Vector2ndarray(Output, FullShape);
}









ndarray GetCoreShellCouplingBoundShell()
{
  using namespace CORESHELL;

  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreIndex.size(),
                                   coreshellSet.ShellMaterial.size(),
                                   coreshellSet.nMedium.size(),
                                   detectorSet.NA.size(),
                                   detectorSet.PhiOffset.size(),
                                   detectorSet.GammaOffset.size(),
                                   detectorSet.Filter.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;


  DVector Output(FullSize);

  #pragma omp parallel for collapse(12)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                                  for (size_t na=0; na<FullShape[8]; ++na)
                                      for (size_t p=0; p<FullShape[9]; ++p)
                                          for (size_t g=0; g<FullShape[10]; ++g)
                                              for (size_t f=0; f<FullShape[11]; ++f)
                                                  {
                                                    size_t idx = f  +
                                                                 g  * FullShape[11] +
                                                                 p  * FullShape[11] * FullShape[10] +
                                                                 na * FullShape[11] * FullShape[10] * FullShape[9] +
                                                                 n  * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] +
                                                                 Si * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] +
                                                                 Ci * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] +
                                                                 Sd * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] +
                                                                 Cd * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                                                 a  * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                                                 j  * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                                                 w  * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                                                 ;

                                                    SOURCE::State    sourceState     = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                                    CORESHELL::State scattererState  = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreIndex[Ci], coreshellSet.ShellMaterial[Si][w], coreshellSet.nMedium[n]);
                                                    DETECTOR::State  detectorState   = DETECTOR::State(detectorSet.ScalarField[0], detectorSet.NA[na], detectorSet.PhiOffset[p], detectorSet.GammaOffset[g], detectorSet.Filter[f], detectorSet.Coherent, detectorSet.PointCoupling);

                                                    CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState);
                                                    DETECTOR::Detector det = DETECTOR::Detector(detectorState);

                                                    Output[idx] = abs( det.Coupling(Scat) );
                                                  }

  return Vector2ndarray(Output, FullShape);
}





ndarray GetCoreShellCouplingBoundCoreBoundShell()
{
  using namespace CORESHELL;
  std::vector<size_t> FullShape = {sourceSet.Wavelength.size(),
                                   sourceSet.Jones.size(),
                                   sourceSet.Amplitude.size(),
                                   coreshellSet.CoreDiameter.size(),
                                   coreshellSet.ShellDiameter.size(),
                                   coreshellSet.CoreMaterial.size(),
                                   coreshellSet.ShellMaterial.size(),
                                   coreshellSet.nMedium.size(),
                                   detectorSet.NA.size(),
                                   detectorSet.PhiOffset.size(),
                                   detectorSet.GammaOffset.size(),
                                   detectorSet.Filter.size()};

  size_t FullSize = 1;
  for (auto e: FullShape)
      FullSize *= e;

  DVector Output(FullSize);

  #pragma omp parallel for collapse(12)
  for (size_t w=0; w<FullShape[0]; ++w)
      for (size_t j=0; j<FullShape[1]; ++j)
          for (size_t a=0; a<FullShape[2]; ++a)
              for (size_t Cd=0; Cd<FullShape[3]; ++Cd)
                  for (size_t Sd=0; Sd<FullShape[4]; ++Sd)
                      for (size_t Ci=0; Ci<FullShape[5]; ++Ci)
                          for (size_t Si=0; Si<FullShape[6]; ++Si)
                              for (size_t n=0; n<FullShape[7]; ++n)
                                  for (size_t na=0; na<FullShape[8]; ++na)
                                      for (size_t p=0; p<FullShape[9]; ++p)
                                          for (size_t g=0; g<FullShape[10]; ++g)
                                              for (size_t f=0; f<FullShape[11]; ++f)
                                                  {
                                                    size_t idx = f  +
                                                                 g  * FullShape[11] +
                                                                 p  * FullShape[11] * FullShape[10] +
                                                                 na * FullShape[11] * FullShape[10] * FullShape[9] +
                                                                 n  * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] +
                                                                 Si * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] +
                                                                 Ci * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] +
                                                                 Sd * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] +
                                                                 Cd * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] +
                                                                 a  * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] +
                                                                 j  * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] +
                                                                 w  * FullShape[11] * FullShape[10] * FullShape[9] * FullShape[8] * FullShape[7] * FullShape[6] * FullShape[5] * FullShape[4] * FullShape[3] * FullShape[2] * FullShape[1];
                                                                 ;

                                                    SOURCE::State    sourceState     = SOURCE::State(sourceSet.Wavelength[w], sourceSet.Jones[j], sourceSet.Amplitude[a]);
                                                    CORESHELL::State scattererState  = CORESHELL::State(coreshellSet.CoreDiameter[Cd], coreshellSet.ShellDiameter[Sd], coreshellSet.CoreMaterial[Ci][w], coreshellSet.ShellMaterial[Si][w], coreshellSet.nMedium[n]);
                                                    DETECTOR::State  detectorState   = DETECTOR::State(detectorSet.ScalarField[0], detectorSet.NA[na], detectorSet.PhiOffset[p], detectorSet.GammaOffset[g], detectorSet.Filter[f], detectorSet.Coherent, detectorSet.PointCoupling);

                                                    CORESHELL::Scatterer Scat = CORESHELL::Scatterer(scattererState, sourceState);
                                                    DETECTOR::Detector det = DETECTOR::Detector(detectorState);

                                                    Output[idx] = abs( det.Coupling(Scat) );
                                                  }

  return Vector2ndarray(Output, FullShape);
}












ndarray GetSphereQsca()    { return GetSphereData( &SPHERE::Scatterer::GetQsca ) ; }
ndarray GetSphereQext()    { return GetSphereData( &SPHERE::Scatterer::GetQext ) ; }
ndarray GetSphereQabs()    { return GetSphereData( &SPHERE::Scatterer::GetQabs ) ; }
ndarray GetSphereQpr()     { return GetSphereData( &SPHERE::Scatterer::GetQpr ) ; }
ndarray GetSphereQback()   { return GetSphereData( &SPHERE::Scatterer::GetQback ) ; }
ndarray GetSphereQforward(){ return GetSphereData( &SPHERE::Scatterer::GetQforward ) ; }
ndarray GetSphereCsca()    { return GetSphereData( &SPHERE::Scatterer::GetCsca ) ; }
ndarray GetSphereCext()    { return GetSphereData( &SPHERE::Scatterer::GetCext ) ; }
ndarray GetSphereCabs()    { return GetSphereData( &SPHERE::Scatterer::GetCabs ) ; }
ndarray GetSphereCpr()     { return GetSphereData( &SPHERE::Scatterer::GetCpr ) ; }
ndarray GetSphereCback()   { return GetSphereData( &SPHERE::Scatterer::GetCback ) ; }
ndarray GetSphereCforward(){ return GetSphereData( &SPHERE::Scatterer::GetCforward ) ; }
ndarray GetSphereG()       { return GetSphereData( &SPHERE::Scatterer::GetG ) ; }

Cndarray GetSphereAn(size_t MaxOrder){ return GetSphereCoefficient( &SPHERE::Scatterer::GetAn, MaxOrder ) ; }
Cndarray GetSphereBn(size_t MaxOrder){ return GetSphereCoefficient( &SPHERE::Scatterer::GetBn, MaxOrder ) ; }
Cndarray GetSphereA1(){ return GetSphereCoefficient( &SPHERE::Scatterer::GetAn, 1 ) ; }
Cndarray GetSphereB1(){ return GetSphereCoefficient( &SPHERE::Scatterer::GetBn, 1 ) ; }
Cndarray GetSphereA2(){ return GetSphereCoefficient( &SPHERE::Scatterer::GetAn, 2 ) ; }
Cndarray GetSphereB2(){ return GetSphereCoefficient( &SPHERE::Scatterer::GetBn, 2 ) ; }
Cndarray GetSphereA3(){ return GetSphereCoefficient( &SPHERE::Scatterer::GetAn, 3 ) ; }
Cndarray GetSphereB3(){ return GetSphereCoefficient( &SPHERE::Scatterer::GetBn, 3 ) ; }


ndarray GetCylinderQsca()    { return GetCylinderData( &CYLINDER::Scatterer::GetQsca ) ; }
ndarray GetCylinderQext()    { return GetCylinderData( &CYLINDER::Scatterer::GetQext ) ; }
ndarray GetCylinderQabs()    { return GetCylinderData( &CYLINDER::Scatterer::GetQabs ) ; }
ndarray GetCylinderQpr()     { return GetCylinderData( &CYLINDER::Scatterer::GetQpr ) ; }
ndarray GetCylinderQback()   { return GetCylinderData( &CYLINDER::Scatterer::GetQback ) ; }
ndarray GetCylinderQforward(){ return GetCylinderData( &CYLINDER::Scatterer::GetQforward ) ; }
ndarray GetCylinderCsca()    { return GetCylinderData( &CYLINDER::Scatterer::GetCsca ) ; }
ndarray GetCylinderCext()    { return GetCylinderData( &CYLINDER::Scatterer::GetCext ) ; }
ndarray GetCylinderCabs()    { return GetCylinderData( &CYLINDER::Scatterer::GetCabs ) ; }
ndarray GetCylinderCpr()     { return GetCylinderData( &CYLINDER::Scatterer::GetCpr ) ; }
ndarray GetCylinderCback()   { return GetCylinderData( &CYLINDER::Scatterer::GetCback ) ; }
ndarray GetCylinderCforward(){ return GetCylinderData( &CYLINDER::Scatterer::GetCforward ) ; }
ndarray GetCylinderG()       { return GetCylinderData( &CYLINDER::Scatterer::GetG ) ; }

Cndarray GetCylinderA1n(size_t MaxOrder){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetA1n, MaxOrder ) ; }
Cndarray GetCylinderB1n(size_t MaxOrder){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetB1n, MaxOrder ) ; }
Cndarray GetCylinderA2n(size_t MaxOrder){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetA2n, MaxOrder ) ; }
Cndarray GetCylinderB2n(size_t MaxOrder){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetB2n, MaxOrder ) ; }
Cndarray GetCylinderA11(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetA1n, 1 ) ; }
Cndarray GetCylinderB11(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetB1n, 1 ) ; }
Cndarray GetCylinderA21(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetA2n, 1 ) ; }
Cndarray GetCylinderB21(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetB2n, 1 ) ; }
Cndarray GetCylinderA12(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetA1n, 2 ) ; }
Cndarray GetCylinderB12(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetB1n, 2 ) ; }
Cndarray GetCylinderA22(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetA2n, 2 ) ; }
Cndarray GetCylinderB22(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetB2n, 2 ) ; }
Cndarray GetCylinderA13(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetA1n, 3 ) ; }
Cndarray GetCylinderB13(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetB1n, 3 ) ; }
Cndarray GetCylinderA23(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetA2n, 3 ) ; }
Cndarray GetCylinderB23(){ return GetCylinderCoefficient( &CYLINDER::Scatterer::GetB2n, 3 ) ; }



ndarray GetCoreShellQsca()    { return GetCoreShellData( &CORESHELL::Scatterer::GetQsca ) ; }
ndarray GetCoreShellQext()    { return GetCoreShellData( &CORESHELL::Scatterer::GetQext ) ; }
ndarray GetCoreShellQabs()    { return GetCoreShellData( &CORESHELL::Scatterer::GetQabs ) ; }
ndarray GetCoreShellQpr()     { return GetCoreShellData( &CORESHELL::Scatterer::GetQpr ) ; }
ndarray GetCoreShellQback()   { return GetCoreShellData( &CORESHELL::Scatterer::GetQback ) ; }
ndarray GetCoreShellQforward(){ return GetCoreShellData( &CORESHELL::Scatterer::GetQforward ) ; }
ndarray GetCoreShellCsca()    { return GetCoreShellData( &CORESHELL::Scatterer::GetCsca ) ; }
ndarray GetCoreShellCext()    { return GetCoreShellData( &CORESHELL::Scatterer::GetCext ) ; }
ndarray GetCoreShellCabs()    { return GetCoreShellData( &CORESHELL::Scatterer::GetCabs ) ; }
ndarray GetCoreShellCpr()     { return GetCoreShellData( &CORESHELL::Scatterer::GetCpr ) ; }
ndarray GetCoreShellCback()   { return GetCoreShellData( &CORESHELL::Scatterer::GetCback ) ; }
ndarray GetCoreShellCforward(){ return GetCoreShellData( &CORESHELL::Scatterer::GetCforward ) ; }
ndarray GetCoreShellG()       { return GetCoreShellData( &CORESHELL::Scatterer::GetG ) ; }

Cndarray GetCoreShellAn(size_t MaxOrder){ return GetCoreShellCoefficient( &CORESHELL::Scatterer::GetAn, MaxOrder ) ; }
Cndarray GetCoreShellBn(size_t MaxOrder){ return GetCoreShellCoefficient( &CORESHELL::Scatterer::GetBn, MaxOrder ) ; }
Cndarray GetCoreShellA1(){ return GetCoreShellCoefficient( &CORESHELL::Scatterer::GetAn, 1 ) ; }
Cndarray GetCoreShellB1(){ return GetCoreShellCoefficient( &CORESHELL::Scatterer::GetBn, 1 ) ; }
Cndarray GetCoreShellA2(){ return GetCoreShellCoefficient( &CORESHELL::Scatterer::GetAn, 2 ) ; }
Cndarray GetCoreShellB2(){ return GetCoreShellCoefficient( &CORESHELL::Scatterer::GetBn, 2 ) ; }
Cndarray GetCoreShellA3(){ return GetCoreShellCoefficient( &CORESHELL::Scatterer::GetAn, 3 ) ; }
Cndarray GetCoreShellB3(){ return GetCoreShellCoefficient( &CORESHELL::Scatterer::GetBn, 3 ) ; }
};


PYBIND11_MODULE(Experiment, module)
{
  py::class_<Experiment>(module, "CppExperiment")
  .def(py::init<>())

  .def("SetDetector", &Experiment::SetDetector,  py::arg("CppDetectorSet") )

  .def("SetSource", &Experiment::SetSource,  py::arg("CppSourceSet") )

  .def("SetSphere", &Experiment::SetSphere, py::arg("CppSphereSet") )

  .def("SetCylinder", &Experiment::SetCylinder, py::arg("CppCylinderSet") )

  .def("SetCoreShell", &Experiment::SetCoreShell, py::arg("CppCoreShellSet") )

  .def("GetSphereQsca",            &Experiment::GetSphereQsca)
  .def("GetSphereQext",            &Experiment::GetSphereQext)
  .def("GetSphereQabs",            &Experiment::GetSphereQabs)
  .def("GetSphereQabs",            &Experiment::GetSphereQabs)
  .def("GetSphereQpr",             &Experiment::GetSphereQpr)
  .def("GetSphereQforward",        &Experiment::GetSphereQforward)
  .def("GetSphereQback",           &Experiment::GetSphereQback)
  .def("GetSphereCsca",            &Experiment::GetSphereCsca)
  .def("GetSphereCext",            &Experiment::GetSphereCext)
  .def("GetSphereCabs",            &Experiment::GetSphereCabs)
  .def("GetSphereCabs",            &Experiment::GetSphereCabs)
  .def("GetSphereCpr",             &Experiment::GetSphereCpr)
  .def("GetSphereCforward",        &Experiment::GetSphereCforward)
  .def("GetSphereCback",           &Experiment::GetSphereCback)
  .def("GetSphereCoupling",        &Experiment::GetSphereCoupling)

  .def("GetSphereAn",              &Experiment::GetSphereAn)
  .def("GetSphereBn",              &Experiment::GetSphereBn)
  .def("GetSphereA1",              &Experiment::GetSphereA1)
  .def("GetSphereB1",              &Experiment::GetSphereB1)
  .def("GetSphereA2",              &Experiment::GetSphereA2)
  .def("GetSphereB2",              &Experiment::GetSphereB2)


  .def("GetCylinderQsca",          &Experiment::GetCylinderQsca)
  .def("GetCylinderQext",          &Experiment::GetCylinderQext)
  .def("GetCylinderQabs",          &Experiment::GetCylinderQabs)
  .def("GetCylinderQabs",          &Experiment::GetCylinderQabs)
  .def("GetCylinderQpr",           &Experiment::GetCylinderQpr)
  .def("GetCylinderQforward",      &Experiment::GetCylinderQforward)
  .def("GetCylinderQback",         &Experiment::GetCylinderQback)
  .def("GetCylinderCsca",          &Experiment::GetCylinderCsca)
  .def("GetCylinderCext",          &Experiment::GetCylinderCext)
  .def("GetCylinderCabs",          &Experiment::GetCylinderCabs)
  .def("GetCylinderCabs",          &Experiment::GetCylinderCabs)
  .def("GetCylinderCpr",           &Experiment::GetCylinderCpr)
  .def("GetCylinderCforward",      &Experiment::GetCylinderCforward)
  .def("GetCylinderCback",         &Experiment::GetCylinderCback)
  .def("GetCylinderCoupling",      &Experiment::GetCylinderCoupling)

  .def("GetCylinderA1n",           &Experiment::GetCylinderA1n)
  .def("GetCylinderA2n",           &Experiment::GetCylinderA2n)
  .def("GetCylinderB1n",           &Experiment::GetCylinderB1n)
  .def("GetCylinderB2n",           &Experiment::GetCylinderB2n)
  .def("GetCylinderA11",           &Experiment::GetCylinderA11)
  .def("GetCylinderA21",           &Experiment::GetCylinderA21)
  .def("GetCylinderB11",           &Experiment::GetCylinderB11)
  .def("GetCylinderB21",           &Experiment::GetCylinderB21)
  .def("GetCylinderA12",           &Experiment::GetCylinderA12)
  .def("GetCylinderA22",           &Experiment::GetCylinderA22)
  .def("GetCylinderB12",           &Experiment::GetCylinderB12)
  .def("GetCylinderB22",           &Experiment::GetCylinderB22)
  .def("GetCylinderA13",           &Experiment::GetCylinderA13)
  .def("GetCylinderA23",           &Experiment::GetCylinderA23)
  .def("GetCylinderB13",           &Experiment::GetCylinderB13)
  .def("GetCylinderB23",           &Experiment::GetCylinderB23)


  .def("GetCoreShellQsca",         &Experiment::GetCoreShellQsca)
  .def("GetCoreShellQext",         &Experiment::GetCoreShellQext)
  .def("GetCoreShellQabs",         &Experiment::GetCoreShellQabs)
  .def("GetCoreShellQabs",         &Experiment::GetCoreShellQabs)
  .def("GetCoreShellQpr",          &Experiment::GetCoreShellQpr)
  .def("GetCoreShellQforward",     &Experiment::GetCoreShellQforward)
  .def("GetCoreShellQback",        &Experiment::GetCoreShellQback)
  .def("GetCoreShellCsca",         &Experiment::GetCoreShellCsca)
  .def("GetCoreShellCext",         &Experiment::GetCoreShellCext)
  .def("GetCoreShellCabs",         &Experiment::GetCoreShellCabs)
  .def("GetCoreShellCabs",         &Experiment::GetCoreShellCabs)
  .def("GetCoreShellCpr",          &Experiment::GetCoreShellCpr)
  .def("GetCoreShellCforward",     &Experiment::GetCoreShellCforward)
  .def("GetCoreShellCback",        &Experiment::GetCoreShellCback)
  .def("GetCoreShellCoupling",     &Experiment::GetCoreShellCoupling)

  .def("GetCoreShellAn",           &Experiment::GetCoreShellAn)
  .def("GetCoreShellBn",           &Experiment::GetCoreShellBn)
  .def("GetCoreShellA1",           &Experiment::GetCoreShellA1)
  .def("GetCoreShellB1",           &Experiment::GetCoreShellB1)
  .def("GetCoreShellA2",           &Experiment::GetCoreShellA2)
  .def("GetCoreShellB2",           &Experiment::GetCoreShellB2)
  .def("GetCoreShellA3",           &Experiment::GetCoreShellA3)
  .def("GetCoreShellB3",           &Experiment::GetCoreShellB3)

  ;

}
