#include "includes/pybind11.hpp"
#include "Sphere.cpp"
#include "CoreShell.cpp"
#include "Cylinder.cpp"
#include "Detectors.cpp"
#include "Experiment.cpp"

#include <iostream>
#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>

namespace py=pybind11;

PYBIND11_MODULE(LMTScatterer, module)
{
    module.doc() = "Lorenz-Mie Theory (LMT) C++ binding module for PyMieSim Python package.";


      pybind11::class_<SPHERE::State>(module,    "CppSphereState");
      pybind11::class_<CORESHELL::State>(module, "CppCoreShellState");
      pybind11::class_<CYLINDER::State>(module,  "CppCylinderState");


      py::class_<SPHERE::Scatterer>(module, "SPHERE")
      .def(py::init<double&, double&, double&, complex128&, double&, CVector&>(),
           py::arg("Wavelength"),
           py::arg("Amplitude"),
           py::arg("Diameter"),
           py::arg("Index"),
           py::arg("nMedium"),
           py::arg("Jones")
          )

      .def("S1S2",
      &SPHERE::Scatterer::PyS1S2_,
      py::arg("Phi") )

      .def("Fields",
      &SPHERE::Scatterer::PyFields_,
      py::arg("Phi"),
      py::arg("Theta"),
      py::arg("R") )

      .def("FullFields",
      &SPHERE::Scatterer::PyFullFields_,
      py::arg("Sampling"),
      py::arg("R") )


      .def("an", py::overload_cast<>(&SPHERE::Scatterer::GetPyAn))
      .def("bn", py::overload_cast<>(&SPHERE::Scatterer::GetPyBn))
      .def("cn", py::overload_cast<>(&SPHERE::Scatterer::GetPyCn))
      .def("dn", py::overload_cast<>(&SPHERE::Scatterer::GetPyDn))

      .def_property_readonly("Qsca",  &SPHERE::Scatterer::GetQsca)
      .def_property_readonly("Qext",  &SPHERE::Scatterer::GetQext)
      .def_property_readonly("Qabs",  &SPHERE::Scatterer::GetQabs)
      .def_property_readonly("Qback", &SPHERE::Scatterer::GetQback)
      .def_property_readonly("Qpr",   &SPHERE::Scatterer::GetQpr)


      .def_property_readonly("Csca",  &SPHERE::Scatterer::GetCsca)
      .def_property_readonly("Cext",  &SPHERE::Scatterer::GetCext)
      .def_property_readonly("Cabs",  &SPHERE::Scatterer::GetCabs)
      .def_property_readonly("Cback", &SPHERE::Scatterer::GetCback)
      .def_property_readonly("Cpr",   &SPHERE::Scatterer::GetCpr)

      .def_property_readonly("g",     &SPHERE::Scatterer::GetG)


      .def_readwrite("State", &SPHERE::Scatterer::state)
      .def_readwrite("Area", &SPHERE::Scatterer::Area)
      .def_readwrite("SizeParameter", &SPHERE::Scatterer::SizeParameter)
      ;


      py::class_<CORESHELL::Scatterer>(module, "CORESHELL")
      .def(py::init<double &, double &, double &, double &, complex128 &, complex128 &, double &, CVector &>(),
           py::arg("Wavelength"),
           py::arg("Amplitude"),
           py::arg("CoreDiameter"),
           py::arg("ShellDiameter"),
           py::arg("CoreIndex"),
           py::arg("ShellIndex"),
           py::arg("nMedium"),
           py::arg("Jones")
          )

       .def("S1S2",
            &CORESHELL::Scatterer::PyS1S2_,
            py::arg("Phi") )

      .def("Fields",
           &CORESHELL::Scatterer::PyFields_,
           py::arg("Phi"),
           py::arg("Theta"),
           py::arg("R") )

      .def("FullFields",
           &CORESHELL::Scatterer::PyFullFields_,
           py::arg("Sampling"),
           py::arg("R") )

      .def("an", py::overload_cast<>(&CORESHELL::Scatterer::GetPyAn))
      .def("bn", py::overload_cast<>(&CORESHELL::Scatterer::GetPyBn))

      .def_property_readonly("Qsca",  &CORESHELL::Scatterer::GetQsca)
      .def_property_readonly("Qext",  &CORESHELL::Scatterer::GetQext)
      .def_property_readonly("Qabs",  &CORESHELL::Scatterer::GetQabs)
      .def_property_readonly("Qback", &CORESHELL::Scatterer::GetQback)
      .def_property_readonly("Qpr",   &CORESHELL::Scatterer::GetQpr)

      .def_property_readonly("Csca",  &CORESHELL::Scatterer::GetCsca)
      .def_property_readonly("Cext",  &CORESHELL::Scatterer::GetCext)
      .def_property_readonly("Cabs",  &CORESHELL::Scatterer::GetCabs)
      .def_property_readonly("Cback", &CORESHELL::Scatterer::GetCback)
      .def_property_readonly("Cpr",   &CORESHELL::Scatterer::GetCpr)

      .def_property_readonly("g",     &CORESHELL::Scatterer::GetG)

      .def_readwrite("State", &CORESHELL::Scatterer::state)
      .def_readwrite("Area", &CORESHELL::Scatterer::Area)
      .def_readwrite("SizeParameter", &CORESHELL::Scatterer::SizeParameter)
      ;






      py::class_<CYLINDER::Scatterer>(module, "CYLINDER")
      .def(py::init<double&, double&, double&, complex128&, double&, CVector&>(),
           py::arg("Wavelength"),
           py::arg("Amplitude"),
           py::arg("Diameter"),
           py::arg("Index"),
           py::arg("nMedium"),
           py::arg("Jones")
          )

       .def("S1S2",
            &CYLINDER::Scatterer::PyS1S2_,
            py::arg("Phi") )

      .def("Fields",
           &CYLINDER::Scatterer::PyFields_,
           py::arg("Phi"),
           py::arg("Theta"),
           py::arg("R") )

      .def("FullFields",
           &CYLINDER::Scatterer::PyFullFields_,
           py::arg("Sampling"),
           py::arg("R") )

      .def("a1n", py::overload_cast<>(&CYLINDER::Scatterer::GetPyA1n))
      .def("b1n", py::overload_cast<>(&CYLINDER::Scatterer::GetPyB1n))
      .def("a2n", py::overload_cast<>(&CYLINDER::Scatterer::GetPyA2n))
      .def("b2n", py::overload_cast<>(&CYLINDER::Scatterer::GetPyB2n))

      .def_property_readonly("Qsca",  &CYLINDER::Scatterer::GetQsca)
      .def_property_readonly("Qext",  &CYLINDER::Scatterer::GetQext)
      .def_property_readonly("Qabs",  &CYLINDER::Scatterer::GetQabs)
      .def_property_readonly("Qback", &CYLINDER::Scatterer::GetQback)
      .def_property_readonly("Qpr",   &CYLINDER::Scatterer::GetQpr)

      .def_property_readonly("Csca",  &CYLINDER::Scatterer::GetCsca)
      .def_property_readonly("Cext",  &CYLINDER::Scatterer::GetCext)
      .def_property_readonly("Cabs",  &CYLINDER::Scatterer::GetCabs)
      .def_property_readonly("Cback", &CYLINDER::Scatterer::GetCback)
      .def_property_readonly("Cpr",   &CYLINDER::Scatterer::GetCpr)

      .def_property_readonly("g",     &CYLINDER::Scatterer::GetG)

      .def_readwrite("State", &CYLINDER::Scatterer::state)
      .def_readwrite("Area", &CYLINDER::Scatterer::Area)
      .def_readwrite("SizeParameter", &CYLINDER::Scatterer::SizeParameter)
      ;


     

      py::class_<DETECTOR::Detector>(module, "BindedDetector")
      .def(py::init<CVector&, double &, double &, double &, double &, bool &, bool &>(),
           py::arg("ScalarField"),
           py::arg("NA"),
           py::arg("PhiOffset"),
           py::arg("GammaOffset"),
           py::arg("Filter"),
           py::arg("Coherent"),
           py::arg("PointCoupling")  //true = point ; false=mean
          )

      .def(py::init<CVector &, double &, double &, double &, double &, bool &, bool &>(),
           py::arg("ScalarField"),
           py::arg("NA"),
           py::arg("PhiOffset"),
           py::arg("GammaOffset"),
           py::arg("Filter"),
           py::arg("Coherent"),
           py::arg("PointCoupling")  //true = point ; false=mean
          )

       .def("CouplingSphere",    &DETECTOR::Detector::Coupling<SPHERE::Scatterer>,    py::arg("Scatterer") )
       .def("CouplingCylinder",  &DETECTOR::Detector::Coupling<CYLINDER::Scatterer>,  py::arg("Scatterer") )
       .def("CouplingCoreShell", &DETECTOR::Detector::Coupling<CORESHELL::Scatterer>, py::arg("Scatterer") )
       ;
}







// -
