#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dataclasses import dataclass
import numpy

from PyMieSim.Tools.Polarization import LinearPolarization, JonesVector, RightCircularPolarization, LeftCircularPolarization
from MPSPlots.Utils import ToList


@dataclass
class PlaneWave():
    """
    .. note::
        Class representing plane wave beam as a light source for
        light scattering.
    """

    Wavelength: float 
    """ Wavelenght of the light field. """
    Polarization: float = 0
    """ Polarization of the light field in degree. """
    Amplitude: float = 1
    """ Maximal value of the electric field at focus point. """

    def __post_init__(self):
        self.k = 2 * numpy.pi / self.Wavelength
        self.Amplitude = ToList(self.Amplitude, dtype=numpy.float)
        self.Polarization = ToList(self.Polarization, dtype=numpy.float)
        self.Wavelength = ToList(self.Wavelength, dtype=numpy.float)

        if isinstance(self.Polarization, (LinearPolarization, RightCircularPolarization, LeftCircularPolarization, JonesVector)):
            self.Polarization = self.Polarization
        else:
            self.Polarization = LinearPolarization(Angle=self.Polarization)
