#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy
from dataclasses import dataclass

from MPSPlots.Math import Angle as AngleCls
from MPSPlots.Utils import ToList


class Polarization():
    def __init__(self):
        pass


@dataclass
class JonesVector(Polarization):
    Description = 'Jones vector'

    def __init__(self, J: list):
        self.J = numpy.asarray(J).astype(complex)
        #self.ComputeParameters()

    @property
    def Size(self):
        return self.J.shape[0]

    def ComputeParameters(self):
        """relation taken from
        http://optics.hanyang.ac.kr/~choh/degree/[2014-1]%20photonics_graduated/Polarization%20Optics(ch6)new.pdf
        slide 6"""

        self.Delay = _Angle(numpy.angle(self.J[0]) - numpy.angle(self.J[1]), Unit='Radian')

        if abs(self.J[1]) == 0:
            r = 1e6
            arg0 = 2 * r / (1 - numpy.abs(r)**2) * numpy.cos(self.Delay.Radian)
            arg1 = 2 * r / (1 + numpy.abs(r)**2) * numpy.sin(self.Delay.Radian)
            self.Azimuth = AngleCls(numpy.arctan(abs(arg0)) / 2, Unit='Radian')
            self.Ellipsity = AngleCls(numpy.arcsin(abs(arg1)) / 2, Unit='Radian')

        elif abs(self.J[0]) == abs(self.J[1]):
            r = abs(self.J[0]) / abs(self.J[1])
            arg0 = numpy.cos(self.Delay.Radian) * 1e6
            arg1 = numpy.sin(self.Delay.Radian)
            self.Azimuth = AngleCls(numpy.arctan(arg0) / 2, Unit='Radian')
            self.Ellipsity = AngleCls(numpy.arcsin(arg1) / 2, Unit='Radian')

        else:
            r = abs(self.J[0]) / abs(self.J[1])
            arg0 = 2 * r / (1 - numpy.abs(r)**2) * numpy.cos(self.Delay.Radian)
            arg1 = 2 * r / (1 + numpy.abs(r)**2) * numpy.sin(self.Delay.Radian)
            self.Azimuth = AngleCls(numpy.arctan(numpy.abs(arg0)) / 2, Unit='Radian')
            self.Ellipsity = AngleCls(numpy.arcsin(numpy.abs(arg1)) / 2, Unit='Radian')


@dataclass
class LinearPolarization(JonesVector):
    Description = 'Linear polarization'
    Angle: list
    Unit: str = "Degree"

    def __post_init__(self):
        self.Angle = ToList(self.Angle, dtype=numpy.float)

        if self.Unit.lower() == "degree":
            self.Angle = numpy.deg2rad(self.Angle)  
        else:
            numpy.asarray(self.Angle) 

        J = [numpy.cos(self.Angle), numpy.sin(self.Angle)]

        super().__init__(J=J)


class RightCircularPolarization(JonesVector):
    Description = 'Right circular polarization'

    def __init__(self):
        super().__init__(J=[1, 1j])


class LeftCircularPolarization(JonesVector):
    Description = 'Left circular polarization'

    def __init__(self):
        super().__init__(J=[1, -1j])


#- 
