
import unittest
from pyface.api        import GUI
from PyMieSim.Detector import LPmode, Photodiode, IntegratingSphere
from .utils            import *
import inspect

PLOTTIME = 1000 # one seconde

class DetectorTestCase(unittest.TestCase, BaseStepTest):

    @TestFactory('Photodiode initialisation')
    def step00(self):
        print("\nUnittest: <Detector>" + '='*100)
        self._Photodiode = Photodiode(NA=0.2, Sampling=100, GammaOffset=0, PhiOffset=0)


    @TestFactory('LPmode initialisation')
    def step01(self):
        self._LPmode = LPmode(Mode="0-1", NA=0.1, Sampling=200, GammaOffset=0, PhiOffset=0)


    @TestFactory('IntegratingSphere initialisation')
    def step02(self):
        self._IntegratingSphere = IntegratingSphere(Sampling=300)


    @TestFactory('Photodiode plotting')
    def step03(self):
        GUI.invoke_after(PLOTTIME, Close)
        self._Photodiode.Plot().Show()


    @TestFactory('Photodiode plotting')
    def step04(self):
        GUI.invoke_after(PLOTTIME, Close)
        self._LPmode.Plot().Show()


    @TestFactory('IntegratingSphere plotting')
    def step05(self):
        GUI.invoke_after(PLOTTIME, Close)
        self._IntegratingSphere.Plot().Show()



def suite():
    suite = unittest.TestSuite()
    suite.addTest(DetectorTestCase('test_steps'))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite())
