
import unittest
from pyface.api import GUI
import numpy as np

from PyMieSim.Experiment import SphereSet, Setup, SourceSet, PhotodiodeSet, LPModeSet, CylinderSet, CoreShellSet
from PyMieSim.Materials import *
from .utils import *
from PyMieSim import Measure, Variables

PLOTTIME = 3000 # one seconde

scatSet = CoreShellSet(CoreDiameter=100e-9, ShellDiameter=200e-9, CoreIndex=1.3, ShellIndex=1.4, nMedium=1)
sourceSet = SourceSet(Wavelength=np.linspace(400e-9, 1000e-9, 100), Polarization=0, Amplitude=1)
DetecSet1 = LPModeSet(Mode="0-1", NA=0.3, PhiOffset=0, GammaOffset=0, Filter=None, CouplingMode="Point", Sampling=200)
ExpSet2 = Setup(ScattererSet=scatSet, SourceSet=sourceSet, DetectorSet=DetecSet1)


class ExperimentTestCase(unittest.TestCase, BaseStepTest):

    #___________________________________SPHERE________________________________________________________________
    @TestFactory('Initializing coefficient experiment')
    def step00(self):
        print("\nUnittest: <Experiment>" + '='*100)
        scatSet      = SphereSet(Diameter = [200e-9], Index = [4], nMedium = [1])
        sourceSet    = SourceSet(Wavelength = np.linspace(400e-9, 1000e-9, 100), Polarization = 0, Amplitude = 1)
        self.ExpSet0 = Setup(ScattererSet = scatSet, SourceSet = sourceSet)

    @TestFactory('Experiment Efficiencies computing')
    def step01(self):
        self.Data0 = self.ExpSet0.Get([Measure.Qsca])

    @TestFactory('DataVisual Array Efficiencies plotting')
    def step02(self):
        GUI.invoke_after(PLOTTIME, Close)
        self.Data0.Plot(y=Measure.Qsca, x=Variables.Wavelength).Show()

    #___________________________________CYLINDER________________________________________________________________
    @TestFactory('Initializing photodiode coupling experiment')
    def step03(self):
        scatSet      = CylinderSet(Diameter = 200e-9, Index = 4, nMedium = 1)
        sourceSet    = SourceSet(Wavelength = np.linspace(400e-9, 1000e-9, 100), Polarization = 0, Amplitude = 1)
        DetecSet1    = PhotodiodeSet(NA = 0.1, PhiOffset = 30, GammaOffset = 0, Filter = None, CouplingMode = "Point", Sampling = 300)
        self.ExpSet1 = Setup(ScattererSet = scatSet, SourceSet = sourceSet, DetectorSet  = DetecSet1)

    @TestFactory('Experiment photodiode Coupling computing')
    def step04(self):
        self.Data1 = self.ExpSet1.Get(Measure.Coupling)


    @TestFactory('DataVisual Array photodiode Coupling plotting')
    def step05(self):
        GUI.invoke_after(PLOTTIME, Close)
        self.Data1.Plot(y=Measure.Coupling, x=Variables.Wavelength).Show()


    #___________________________________CORESHELL________________________________________________________________
    @TestFactory('Initializing LP mode coupling experiment')
    def step06(self):
        scatSet      = CoreShellSet(CoreDiameter = 100e-9, ShellDiameter = 200e-9, CoreIndex = 1.3, ShellIndex=1.4, nMedium = 1)
        sourceSet    = SourceSet(Wavelength = np.linspace(400e-9, 1000e-9, 100), Polarization = 0, Amplitude = 1)
        DetecSet1    = LPModeSet(Mode="0-1", NA=0.3, PhiOffset = 0, GammaOffset = 0, Filter = None, CouplingMode = "Point", Sampling = 200)
        self.ExpSet2 = Setup(ScattererSet = scatSet, SourceSet = sourceSet, DetectorSet  = DetecSet1)


    @TestFactory('Experiment LP mode Coupling computing')
    def step07(self):
        self.Data2 = self.ExpSet2.Get(Measure.Coupling)


    @TestFactory('DataVisual Array photodiode Coupling plotting')
    def step08(self):
        GUI.invoke_after(PLOTTIME, Close)
        self.Data2.Plot(y=Measure.Coupling, x=Variables.Wavelength).Show()


    #___________________________________SPHERE_MATERIAL________________________________________________________________
    @TestFactory('Initializing coefficient experiment [Sphere 1 material]')
    def step09(self):
        scatSet      = SphereSet(Diameter = 200e-9, Material = BK7, nMedium = 1)
        sourceSet    = SourceSet(Wavelength = np.linspace(400e-9, 1000e-9, 100), Polarization = 0, Amplitude = 1)
        self.ExpSet  = Setup(ScattererSet = scatSet, SourceSet = sourceSet)


    @TestFactory('Experiment Efficiencies computing [Sphere 1 material]')
    def step10(self):
        self.Data = self.ExpSet.Get(Measure.Qsca)


    @TestFactory('DataVisual Array Efficiencies plotting [Sphere 1 material]')
    def step11(self):
        GUI.invoke_after(PLOTTIME, Close)
        self.Data.Plot(y=Measure.Qsca, x=Variables.Wavelength).Show()


    #___________________________________SPHERE_2_MATERIAL________________________________________________________________
    @TestFactory('Initializing coefficient experiment [Sphere 2 material]')
    def step12(self):
        scatSet      = SphereSet(Diameter = 200e-9, Material = [BK7, SI], nMedium = 1)
        sourceSet    = SourceSet(Wavelength = np.linspace(600e-9, 800e-9, 100), Polarization = 0, Amplitude = 1)
        self.ExpSet  = Setup(ScattererSet = scatSet, SourceSet = sourceSet)

    @TestFactory('Experiment Efficiencies computing [Sphere 2 material]')
    def step13(self):
        self.Data = self.ExpSet.Get(Measure.Qsca)

    @TestFactory('DataVisual Array Efficiencies plotting [Sphere 2 material]')
    def step14(self):
        GUI.invoke_after(PLOTTIME, Close)
        self.Data.Plot(y=Measure.Qsca, x=Variables.Wavelength).Show()


    #___________________________________CYLINDER_2_MATERIAL________________________________________________________________
    @TestFactory('Initializing coefficient experiment [Cylinder 2 material]')
    def step15(self):
        scatSet      = CylinderSet(Diameter = 200e-9, Material = [BK7, SI], nMedium = 1)
        sourceSet    = SourceSet(Wavelength = np.linspace(600e-9, 800e-9, 100), Polarization = 0, Amplitude = 1)
        self.ExpSet  = Setup(ScattererSet = scatSet, SourceSet = sourceSet)

    @TestFactory('Experiment Efficiencies computing [Cylinder 2 material]')
    def step16(self):
        self.Data = self.ExpSet.Get(Measure.Qsca)

    @TestFactory('DataVisual Array Efficiencies plotting [Cylinder 2 material]')
    def step17(self):
        GUI.invoke_after(PLOTTIME, Close)
        self.Data.Plot(y=Measure.Qsca, x=Variables.Wavelength).Show()

    #___________________________________CORE/SHELL_2_MATERIAL________________________________________________________________
    @TestFactory('Initializing coefficient experiment [Core/Shell 2 material]')
    def step18(self):
        scatSet      = CoreShellSet(CoreDiameter = 100e-9, ShellDiameter = 200e-9, CoreMaterial =  [BK7, SI], ShellMaterial= [BK7], nMedium = 1)
        sourceSet    = SourceSet(Wavelength = np.linspace(600e-9, 800e-9, 100), Polarization = 0, Amplitude = 1)
        self.ExpSet  = Setup(ScattererSet = scatSet, SourceSet = sourceSet)

    @TestFactory('Experiment Efficiencies computing [Core/Shell 2 material]')
    def step19(self):
        self.Data = self.ExpSet.Get(Measure.Qsca)

    @TestFactory('DataVisual Array Efficiencies plotting [Core/Shell 2 material]')
    def step20(self):
        GUI.invoke_after(PLOTTIME, Close)
        self.Data.Plot(y=Measure.Qsca, x=Variables.Wavelength).Show()



def suite():
    suite = unittest.TestSuite()
    suite.addTest(ExperimentTestCase('test_steps'))

    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite())
